/**
 * Copyright Openmind http://www.openmindonline.it
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package net.sourceforge.openutils.mgnlmedia.media.utils;

import info.magnolia.cms.beans.runtime.FileProperties;
import info.magnolia.cms.core.Content;
import info.magnolia.cms.core.HierarchyManager;
import info.magnolia.cms.core.NodeData;
import info.magnolia.cms.util.ExclusiveWrite;
import info.magnolia.context.MgnlContext;

import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.Transparency;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.Iterator;
import java.util.TimeZone;

import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.stream.MemoryCacheImageOutputStream;
import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;

import net.sourceforge.openutils.mgnlmedia.media.configuration.MediaConfigurationManager;
import net.sourceforge.openutils.mgnlmedia.media.lifecycle.MediaModuleLifecycle;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.ClassUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.math.NumberUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * @author molaschi
 * @version $Id: ImageUtils.java 1004 2009-02-15 22:33:26Z fgiust $
 */
public final class ImageUtils
{

    /**
     * Logger.
     */
    private static Logger log = LoggerFactory.getLogger(ImageUtils.class);

    private static SimpleDateFormat sdf;

    private static final String[] extensions = new String[]{"jpg", "gif", "png" };

    static
    {
        sdf = new SimpleDateFormat();
        sdf.setTimeZone(TimeZone.getTimeZone("GMT"));
    }

    private ImageUtils()
    {
    }

    private static int getType(ColorModel cm)
    {
        if (cm.getTransparency() == Transparency.BITMASK)
        {
            if (cm.isAlphaPremultiplied())
            {
                return BufferedImage.TYPE_4BYTE_ABGR_PRE;
            }
            else
            {
                return BufferedImage.TYPE_4BYTE_ABGR;
            }
        }
        if (cm.getTransparency() == Transparency.TRANSLUCENT)
        {
            if (cm.isAlphaPremultiplied())
            {
                return BufferedImage.TYPE_INT_ARGB_PRE;
            }
            else
            {
                return BufferedImage.TYPE_INT_ARGB;
            }
        }

        if (cm.getPixelSize() == 8)
        {
            return BufferedImage.TYPE_3BYTE_BGR;
        }
        return BufferedImage.TYPE_INT_RGB;
    }

    /**
     * Resize an image to x,y
     * @param original original image
     * @param x new width
     * @param y new height
     * @return resized image
     */
    public static BufferedImage resizeImage(BufferedImage original, int x, int y)
    {
        BufferedImage resizedImage;
        try
        {
            resizedImage = new BufferedImage(x, y, getType(original.getColorModel()));
        }
        catch (NegativeArraySizeException e)
        {
            throw new RuntimeException("NegativeArraySizeException caught when resizing image to ["
                + x
                + ", "
                + y
                + "]");
        }
        Graphics2D graphics2D = resizedImage.createGraphics();
        graphics2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        graphics2D.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);

        if (x > original.getWidth())
        {
            graphics2D.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
            graphics2D.drawImage(original, 0, 0, x, y, null);
        }
        else
        {
            graphics2D.drawImage(original.getScaledInstance(x, y, Image.SCALE_SMOOTH), 0, 0, x, y, null);
        }

        return resizedImage;
    }

    /**
     * Crop an image from left, top for width, height
     * @param original original image
     * @param left start crop point left
     * @param top start crop point top
     * @param width crop width
     * @param height crop height
     * @return resized image
     */
    public static BufferedImage cropImage(BufferedImage original, int left, int top, int width, int height)
    {
        BufferedImage resizedImage = new BufferedImage(width, height, getType(original.getColorModel()));
        Graphics2D graphics2D = resizedImage.createGraphics();
        graphics2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

        if (width > original.getWidth())
        {
            graphics2D.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
        }
        else
        {
            graphics2D.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
        }
        graphics2D.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_DEFAULT);
        graphics2D.drawImage(original, 0, 0, width, height, left, top, width + left, height + top, null);

        return resizedImage;
    }

    /**
     * Resize an image to fit into x,y not changing x/y
     * @param original original image
     * @param x max width
     * @param y max height
     * @return resized image
     */
    public static BufferedImage fitIn(BufferedImage original, int x, int y)
    {
        if (original == null)
        {
            throw new IllegalArgumentException("input image is null");
        }

        float oX = original.getWidth();
        float oY = original.getHeight();

        if (oX < 1 || oY < 1)
        {
            throw new IllegalArgumentException("Broken input image (width=" + oX + ",height=" + oY + ")");
        }

        float oDelta = oX / oY;
        float delta = ((float) x) / ((float) y);

        if (oDelta > delta)
        {
            int newY = (int) (x * oY / oX);
            if (newY < 1)
            {
                newY = 1;
            }
            return resizeImage(original, x, newY);
        }
        else if (oDelta < delta)
        {
            int newX = (int) (y * oX / oY);
            if (newX < 1)
            {
                newX = 1;
            }
            return resizeImage(original, newX, y);
        }
        else
        {
            return resizeImage(original, x, y);
        }
    }

    public static BufferedImage resizeNoCrop(BufferedImage original, int x, int y)
    {
        float oX = original.getWidth();
        float oY = original.getHeight();

        double xRatio = (double) x / oX;
        double yRatio = (double) y / oY;
        double ratio = Math.min(xRatio, yRatio);
        int newX = (int) Math.round(oX * ratio);
        int newY = (int) Math.round(oY * ratio);

        BufferedImage resizedImage = new BufferedImage(x, y, getType(original.getColorModel()));

        Graphics2D graphics2D = resizedImage.createGraphics();
        graphics2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        graphics2D.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);

        if (ratio > 1)
        {
            graphics2D.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
            graphics2D.drawImage(original, (x - newX) / 2, (y - newY) / 2, newX, newY, null);
        }
        else
        {
            graphics2D.drawImage(
                original.getScaledInstance(newX, newY, Image.SCALE_SMOOTH),
                (x - newX) / 2,
                (y - newY) / 2,
                newX,
                newY,
                null);
        }

        return resizedImage;
    }

    /**
     * Resize an image to fill x or y and center and crop what is out
     * @param original original image
     * @param x min width
     * @param y min height
     * @return resized image
     */
    public static BufferedImage fillAndCropCentered(BufferedImage original, int x, int y)
    {
        float oX = original.getWidth();
        float oY = original.getHeight();
        float oDelta = oX / oY;
        float delta = ((float) x) / ((float) y);

        if (oDelta > delta)
        {
            int newX = (int) (y * oX / oY);
            BufferedImage filled = resizeImage(original, newX, y);
            return cropImage(filled, (newX - x) / 2, 0, x, y);
        }
        else if (oDelta < delta)
        {
            int newY = (int) (x * oY / oX);
            BufferedImage filled = resizeImage(original, x, newY);
            return cropImage(filled, 0, (newY - y) / 2, x, y);
        }
        else
        {
            return resizeImage(original, x, y);
        }
    }

    /**
     * Save a resolution for an image to a node (in resolutions/res-[width]x[height]/data.jpg)
     * @param image image to save
     * @param saveTo node to save to
     * @param extension extension
     * @throws RepositoryException exception in jcr operations
     * @throws IOException exception converting image to jpg
     */
    public static void saveResolution(BufferedImage image, Content saveTo, String extension)
        throws RepositoryException, IOException
    {
        saveResolution(image, saveTo, null, extension);
    }

    /**
     * Save a resolution for an image to a node (in resolutions/[name]/data.jpg)
     * @param image image to save
     * @param saveTo node to save to
     * @param name name for this resolution
     * @param extension extension
     * @throws RepositoryException exception in jcr operations
     * @throws IOException exception converting image to jpg
     */
    public static void saveResolution(BufferedImage image, Content saveTo, String name, String extension)
        throws RepositoryException, IOException
    {
        Content resolutions = saveTo.getChildByName("resolutions");
        if (resolutions == null)
        {
            resolutions = saveTo.createContent("resolutions", MediaConfigurationManager.RESOLUTIONS);
        }

        String resolution = name;
        if (resolution == null)
        {
            resolution = "res-" + image.getWidth() + "x" + image.getHeight();
        }

        synchronized (ExclusiveWrite.getInstance())
        {

            if (resolutions.hasNodeData(resolution))
            {
                NodeData nd = resolutions.getNodeData(resolution);
                nd.delete();
            }

            NodeData nd = resolutions.createNodeData(resolution, PropertyType.BINARY);
            nd.setValue(getStream(image, extension));
            String mimetype = "image/" + extension;
            if ("jpg".equals(extension))
            {
                mimetype = "image/jpeg";
            }
            nd.setAttribute(FileProperties.PROPERTY_EXTENSION, extension);
            nd.setAttribute(FileProperties.PROPERTY_FILENAME, "data");
            nd.setAttribute(FileProperties.PROPERTY_CONTENTTYPE, mimetype);
            nd.setAttribute(FileProperties.PROPERTY_LASTMODIFIED, "" + new Date().getTime());
            nd.setAttribute(FileProperties.PROPERTY_WIDTH, "" + image.getWidth());
            nd.setAttribute(FileProperties.PROPERTY_HEIGHT, "" + image.getHeight());
        }
    }

    public static InputStream getStream(BufferedImage image, String extension) throws IOException
    {
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        BufferedOutputStream out = new BufferedOutputStream(bos);

        if ("jpg".equals(extension))
        {
            Iterator<ImageWriter> iter = ImageIO.getImageWritersByFormatName("jpeg");
            ImageWriter writer = iter.next();
            ImageWriteParam iwp = writer.getDefaultWriteParam();
            iwp.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
            iwp.setCompressionQuality(1.0f); // an integer between 0 and 1
            // 1 specifies minimum compression and maximum quality
            MemoryCacheImageOutputStream output = new MemoryCacheImageOutputStream(out);
            writer.setOutput(output);
            IIOImage iioimage = new IIOImage(image, null, null);
            writer.write(null, iioimage, iwp);

        }
        else
        {
            try
            {
                ImageIO.write(image, extension, out);
            }
            catch (IOException ex)
            {
                log.error("Error writing image to buffer", ex);
                throw ex;
            }
        }

        return new ByteArrayInputStream(bos.toByteArray());
    }

    public static boolean checkOrCreateResolution(final Content media, final String resolutionTarget)
    {

        Content resolutions = media.getChildByName("resolutions");

        String resolution = resolutionTarget;

        if (!"thumbnail".equals(resolutionTarget) && !"preview".equals(resolutionTarget))
        {
            resolution = "res-" + resolutionTarget;
        }
        try
        {
            if (resolutions != null && resolutions.hasNodeData(resolution))
            {
                return true;
            }
        }
        catch (RepositoryException e1)
        {
            // go on with res calculation
        }
        // check again if (while waiting) someone has done this resolution
        try
        {
            if (resolutions != null && resolutions.hasNodeData(resolution))
            {
                return true;
            }
        }
        catch (RepositoryException e1)
        {
            // go on with res calculation
        }

        String type = media.getNodeData("type").getString();
        if (StringUtils.equals(type, "other"))
        {
            return false;
        }

        try
        {
            MgnlContext.doInSystemContext(new MgnlContext.SystemContextOperation()
            {

                /**
                 * {@inheritDoc}
                 */
                public void exec()
                {
                    HierarchyManager hm = MgnlContext.getHierarchyManager(MediaModuleLifecycle.REPO);
                    String resolution = resolutionTarget;

                    if ("thumbnail".equals(resolution))
                    {
                        resolution = "<100x100";
                    }
                    if ("preview".equals(resolution))
                    {
                        resolution = "<450x450";
                    }

                    Content node;
                    try
                    {
                        node = hm.getContent(media.getHandle());
                    }
                    catch (RepositoryException e)
                    {
                        throw new RuntimeException(e);
                    }

                    try
                    {
                        if (!node.hasNodeData("image"))
                        {
                            throw new RuntimeException("No \"image\" nodedata for media " + media.getHandle());
                        }
                    }
                    catch (RepositoryException e)
                    {
                        throw new RuntimeException(e);
                    }
                    NodeData image = node.getNodeData("image");
                    String extension = image.getAttribute(FileProperties.PROPERTY_EXTENSION);
                    if (!Arrays.asList(extensions).contains(extension))
                    {
                        try
                        {
                            image.setAttribute(FileProperties.PROPERTY_EXTENSION, "jpg");
                        }
                        catch (RepositoryException e)
                        {
                            throw new RuntimeException(e);
                        }
                        extension = "jpg";
                    }

                    BufferedImage original = createBufferedImage(image);

                    String resolutionName = "res-" + resolution;

                    BufferedImage img = ImageUtils.getImageForResolution(original, resolution);

                    try
                    {
                        if ("thumbnail".equals(resolutionTarget) || "preview".equals(resolutionTarget))
                        {
                            ImageUtils.saveResolution(img, node, resolutionTarget, extension);
                        }
                        else
                        {
                            ImageUtils.saveResolution(img, node, resolutionName, extension);
                        }
                    }
                    catch (RepositoryException e)
                    {
                        throw new RuntimeException(e);
                    }
                    catch (IOException e)
                    {
                        throw new RuntimeException(e);
                    }

                    try
                    {
                        hm.save();
                    }
                    catch (RepositoryException e)
                    {
                        throw new RuntimeException(e);
                    }

                }

            });
        }
        catch (RuntimeException ex)
        {
            log.error(ClassUtils.getShortClassName(ex.getClass())
                + " checking resolution for "
                + media.getHandle()
                + ": "
                + ex.getMessage(), ex);
            return false;
        }

        return true;
    }

/**
     * Get image for a resolution
     * @param original original image
     * @param resolution resolution (if it starts with '<' does a fitIn else a fillAndCropCentered
     * @return new image
     */
    public static BufferedImage getImageForResolution(BufferedImage original, String resolution)
    {

        if (original == null)
        {
            throw new IllegalArgumentException("input image is null");
        }

        BufferedImage img = null;

        if (resolution.startsWith("<"))
        {
            String resx = StringUtils.substringBefore(StringUtils.substringAfter(resolution, "<"), "x");
            String resy = StringUtils.substringAfter(resolution, "x");

            img = fitIn(original, NumberUtils.toInt(resx), NumberUtils.toInt(resy));
        }
        else if (resolution.startsWith("o"))
        {
            String resx = StringUtils.substringBefore(StringUtils.substringAfter(resolution, "o"), "x");
            String resy = StringUtils.substringAfter(resolution, "x");

            img = resizeNoCrop(original, NumberUtils.toInt(resx), NumberUtils.toInt(resy));
        }
        else
        {

            String resx = StringUtils.substringBefore(resolution, "x");
            String resy = StringUtils.substringAfter(resolution, "x");

            if (resx.charAt(0) < '0' || resx.charAt(0) > '9')
            {
                resx = resx.substring(1);
            }

            img = fillAndCropCentered(original, NumberUtils.toInt(resx), NumberUtils.toInt(resy));
        }

        return img;
    }

    public static String getExtension(Content media, String resolution)
    {
        try
        {
            Content resolutions = media.getContent("resolutions");
            NodeData res = resolutions.getNodeData(resolution);
            return res.getAttribute(FileProperties.PROPERTY_EXTENSION);
        }
        catch (RepositoryException ex)
        {
            return "jpg";
        }
    }

    /**
     * @param image
     * @return
     */
    public static BufferedImage createBufferedImage(NodeData image)
    {
        InputStream is = image.getStream();
        try
        {
            return ImageIO.read(is);
        }
        catch (IOException e)
        {
            // CMYK jpeg can't be read be ImageIO
            // Caused by: javax.imageio.IIOException: Unsupported Image Type
            BufferedImage result = JpegUtils.processNonStandardImage(image);

            if (result == null)
            {
                // throw the original exception back
                throw new BadImageFormatException("Unable to handle " + image.getHandle(), e);
            }
            return result;

        }
        finally
        {
            IOUtils.closeQuietly(is);
        }
    }

}
